use crate::prelude::*;

pub(crate) struct Args {
    pub(crate) period_ms: u32,
    pub(crate) port: u16,
}

impl Default for Args {
    fn default() -> Self {
        Self {
            period_ms: 1_000,
            port: 9000,
        }
    }
}

pub(crate) struct App {
    client: Client,
    stream: Framed<TcpStream, LengthDelimitedCodec>,
    timer: Interval,
}

impl App {
    pub(crate) async fn new(args: Args) -> Result<Self> {
        let stream = TcpStream::connect(("127.0.0.1", args.port)).await?;
        let stream = Framed::new(stream, LengthDelimitedCodec::new());
        let mut timer = tokio::time::interval(Duration::from_millis(args.period_ms.into()));
        timer.set_missed_tick_behavior(MissedTickBehavior::Skip);
        Ok(Self {
            client: Default::default(),
            stream,
            timer,
        })
    }

    pub(crate) fn poll_run(&mut self, cx: &mut Context<'_>) -> Poll<Result<()>> {
        let mut stream = pin!(&mut self.stream);
        match <_ as futures_sink::Sink<Bytes>>::poll_ready(stream.as_mut(), cx) {
            Poll::Pending => {}
            Poll::Ready(Err(err)) => {
                return Poll::Ready(
                    Err(err).context("Can't check network write half for readiness"),
                );
            }
            Poll::Ready(Ok(())) => {
                if let Some(frame) = self.client.poll_send() {
                    if let Err(err) = stream.as_mut().start_send(frame) {
                        return Poll::Ready(Err(err).context("stream.start_send"));
                    }
                    tracing::debug!("Started send");
                }
                match <_ as futures_sink::Sink<Bytes>>::poll_flush(stream.as_mut(), cx) {
                    Poll::Pending => {}
                    Poll::Ready(Err(err)) => return Poll::Ready(Err(err).context("poll_flush")),
                    Poll::Ready(Ok(())) => {}
                }
            }
        }

        match stream.as_mut().poll_next(cx) {
            Poll::Pending => {}
            Poll::Ready(None) => return Poll::Ready(Err(anyhow!("Server closed cxn"))),
            Poll::Ready(Some(frame)) => {
                cx.waker().wake_by_ref();
                let frame = frame.context("network framing decode")?;
                if let Err(err) = self.client.handle_frame(frame.into()) {
                    return Poll::Ready(Err(err).context("client.handle_frame"));
                }
            }
        }

        if self.timer.poll_tick(cx).is_ready() {
            cx.waker().wake_by_ref();
            if let Err(err) = self.client.handle_timeout() {
                return Poll::Ready(Err(err).context("client.handle_timeout"));
            }
        }

        Poll::Pending
    }
}

#[derive(Default)]
struct Client {
    outbox: VecDeque<Bytes>,
    sequence: u64,
}

impl Client {
    fn handle_frame(&self, frame: Bytes) -> Result<()> {
        match rmp_serde::from_slice(&frame)? {
            ToClient::ChatLine { id, line } => {
                tracing::info!(?id, ?line);
            }
        }
        Ok(())
    }

    fn handle_timeout(&mut self) -> Result<()> {
        if !self.outbox.is_empty() {
            bail!("Dropped message, outbox is full");
        }

        let msg = ToServer::ChatLine {
            line: "There was a time, in the era of great chaos, when the Earth and the Moon were at war with each other. A daredevil from the Moon piloted a bizarre aircraft. It was feared, and because of its shape, called EINHANDER.".to_string(),
            sequence: self.sequence,
        };
        let frame = rmp_serde::to_vec(&msg)?;
        self.outbox.push_back(frame.into());
        self.sequence += 1;
        Ok(())
    }

    fn poll_send(&mut self) -> Option<Bytes> {
        self.outbox.pop_front()
    }
}
